# This module is automatically created from setup.py
git_sha1 = '25fb4649fdd2dbc93a8e57e601d60180229224ae'
local_modifications = False
version = '2017.11.22'
long_version = 'pyrocko-2017.11.22-25fb4649fdd2dbc93a8e57e601d60180229224ae-2018-01-15_15:42:40'  # noqa
installed_date = '2018-01-15_15:42:40'
